// Test to draw some simple triangles in OpenGL 1.1, in D!
// Copyright (C) 1999-2001 Sean L Palmer (seanpalmer@earthlink.net)
// This code is released without any warranty.  Use at your own risk.

import windows;
import winbase;
import wingdi;
import winuser;
import opengl;

import time;
import stream;
import string;
import c.stdio;
//import c.stdlib;

private:

HINSTANCE hinst;
HWND hwnd;
HGLRC dev;
int bitspixel;
int w,h;

enum
{  // texture id's
  TexTex1,
  TexTotal // number of textures
}

const char[][] texnames =
[
  "Tex1.bmp",
];

void glErrorMsg(char[] s,GLenum code)
{
  printf("%s OpenGL error %d", s, code);
  assert(false);
}

bit CheckError()
{
  GLenum err;
  if ((err=glGetError())!=GL_NO_ERROR)
  {
    glErrorMsg("OpenGL Error:",err);
    return true;
  }
  return false;
}

bit Win32ErrorMsg(char[] s)
{
  LPVOID lpMsgBuf;
  DWORD err=GetLastError();
  printf("%s error code 0x%08x",s,err);
  assert(false);
  return false;
}

struct RGBAPaletteEntry
{
  ubyte r,g,b,a;
}

// bmp file format
struct BMPHeader
{
  ushort  FileType;
  uint    FileSize;
  ushort  Reserved1;
  ushort  Reserved2;
  uint    OffBits;
  uint    Size;
  uint    Width;
  uint    Height;
  ushort  Planes;
  ushort  BitCount;
  uint    Compression;
  uint    SizeImage;
  uint    XPelsPerMeter;
  uint    YPelsPerMeter;
  uint    ClrUsed;
  uint    ClrImportant;
}

struct Bitmap
{
  int w,h,bpp;
  ubyte* data;
  RGBAPaletteEntry* pal;

  void Kill()
  {
    if (data) delete data;
    if (pal) delete pal;
  }
  void MakeData()
  {
    if (data) delete data;
    data=new ubyte[w*h*bpp/8];
  }
  void MakePal()
  {
    if (!pal) pal=new RGBAPaletteEntry[256];
  }
}

// load a bmp & allocate memory for it
Bitmap LoadBmp(char[] filename)
{
  Bitmap bmp;
  BMPHeader b;
  // open file & get size
  File f = new File(filename);
  assert(f.size());

  // load the BMP piecemeal to avoid struct packing issues
  // NOTE: this oughta be done byte-order independent
  f.read(b.FileType);
  f.read(b.FileSize);
  f.read(b.Reserved1);
  f.read(b.Reserved2);
  f.read(b.OffBits);
  f.read(b.Size);
  f.read(b.Width);
  f.read(b.Height);
  f.read(b.Planes);
  f.read(b.BitCount);
  f.read(b.Compression);
  f.read(b.SizeImage);
  f.read(b.XPelsPerMeter);
  f.read(b.YPelsPerMeter);
  f.read(b.ClrUsed);
  f.read(b.ClrImportant);
  bmp.w=b.Width;
  bmp.h=b.Height;
  bmp.bpp=b.Planes*b.BitCount;
  if (bmp.bpp<=8)
  {
    bmp.MakePal();
    f.readExact(bmp.pal,256*4);
  }
  f.position(b.OffBits);
  if (!b.SizeImage) b.SizeImage=bmp.w*bmp.h*bmp.bpp/8;
  bmp.MakeData();
  f.readExact(bmp.data,b.SizeImage);
  return bmp;
}

// load an OpenGL texture from a bmp file
bit InitTexture(int slot,char[] filename)
{
  Bitmap b=LoadBmp(filename);
  if (!b.data) return false;

  // init pointers & allocate space for image
  ubyte[] image = new ubyte[b.w*b.h*3];
  if (!image) return 0;
  ubyte* i = image;
  RGBAPaletteEntry* p = b.pal;
  ubyte* d = b.data;
  int sw=(b.w*b.bpp/8+3)&~3;

  // convert paletted or BGR image to an RGB image
  // BMP's are upside down, but that's what OpenGL wants
  switch (b.bpp)
  {
    case 8:
    {
      for (int y=0; y < b.h; ++y)
      {
        d=b.data+y*sw;
        for (int x=0; x < b.w; ++x)
        {
          RGBAPaletteEntry* pe = &b.pal[*d++];
          *i++ = pe.b;
          *i++ = pe.g;
          *i++ = pe.r;
        }
      }
      break;
    }
    case 24:
    {
      for (int y=0; y < b.h; ++y)
      {
        d=b.data+y*sw;
        for (int x=0; x < b.w; ++x)
        {
          *i++ = d[2];
          *i++ = d[1];
          *i++ = d[0];
          d+=3;
        }
      }
      break;
    }
  }

  // build the texture object
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glBindTexture(GL_TEXTURE_2D,slot);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

  // textures aren't working for some reason, still gotta debug that.
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB5,b.w,b.h,0,GL_RGB,GL_UNSIGNED_BYTE,image);
  //gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB5,b.w,b.h,GL_RGB,GL_UNSIGNED_BYTE,image);

  // free buffers
  b.Kill();
  return true;
}

void DoneTexture(int slot)
{
}

void UseMaterial(int slot, bit blend)
{
  glColor4f(255,255,255,255);

  if (blend)
  {
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE); // disable z-buffer writes
  }
  else
  {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE); // enable z-buffer writes
  }

  if (slot<0)
  {
    glDisable(GL_TEXTURE_2D);
  }
  else
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, slot);
  }
}

void UseMaterial(int slot, float r,float g,float b, float a)
{
  glColor4f(r,g,b,a);

  if (a<=0.999F)
  {
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE); // disable z-buffer writes
  }
  else
  {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE); // enable z-buffer writes
  }

  if (slot<0)
  {
    glDisable(GL_TEXTURE_2D);
  }
  else
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, slot);
  }
}

void DrawTri(float x,float y,float z,float r)
{
  glBegin(GL_TRIANGLES);
  glTexCoord2f(0.5F,1.0F);
  glVertex3f(x  , y-r, z);
  glTexCoord2f(0.0F,0.0F);
  glVertex3f(x+r, y+r, z);
  glTexCoord2f(1.0F,0.0F);
  glVertex3f(x-r, y+r, z);
  glEnd();
}

void SetupProjection()
{
  // set viewing projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0F, 1.0F,
           1.0F,-1.0F,
          -1.0F, 10.0F);
  glScalef(1,1,-1);  // +z into screen
  CheckError();
  glMatrixMode(GL_MODELVIEW);
}

void SetupCamera()
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  CheckError();
}

void Render()
{
  // UseMaterial(TexTex1, .2,1,1, 0.65);
  UseMaterial(TexTex1,false);
  DrawTri(0.15F,0.0F,1.0F, 0.75F);
  UseMaterial(-1, 1.0F,0.0F,1.0F,1.0f);
  DrawTri(-0.15F,0.0F,1.25F, 0.75F);
}

bit InitRender()
{
  if (!hwnd || !dev)
    return false;
  // clear color and depth buffers
  glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
  CheckError();
  return true;
}

void DoneRender()
{
  if (dev && hwnd)
  {
    CheckError();
    HDC hdc=GetDC(hwnd);
    SwapBuffers(hdc); // does glFinish() internally
    ReleaseDC(hwnd, hdc);
    CheckError();
  }
}


bit InitDemo()
{
  printf("InitDemo\n");
  for (int i=0; i<TexTotal; ++i)
  {
    printf("init texture %d\n",i);
    InitTexture(i,texnames[i]);
  }
  // commented this cuz it was crashing DMD
  //GLfloat[4] matAmb = [ 1.0F,1.0F,1.0F, 1.0F ];
  //glLightModelfv(GL_LIGHT_MODEL_AMBIENT,matAmb);
  printf("setting opengl states\n");
  glShadeModel(GL_SMOOTH);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_DITHER);
  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glEnable(GL_DEPTH_TEST);
  glClearColor(0,0,0,1);
  glDisable(GL_NORMALIZE);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  glFrontFace(GL_CW);
  CheckError();
  return true;
}

void DoneDemo()
{
  printf("DoneDemo\n");
  for (int i=TexTotal; --i>=0; )
    DoneTexture(i);
}

void Resized(int width,int height)
{
  // if (!dev) return;
  if (width<0 || height<0)
  {
    if (hwnd)
    {
      RECT r;
      GetClientRect(hwnd,&r);
      width = r.right;
      height= r.bottom;
    }
  }
  // set viewport to cover the window
  w=width; h=height;
  glViewport(0,0,w,h);
}

// The standard ChoosePixelFormat algorithm fails for
// some situations, so I'm using a replacement algorithm
// that, while it may not be better for all situations,
// at least it works for the ones I care about.
int ChoosePixelFormat2(HDC dc, in PIXELFORMATDESCRIPTOR pfd)
{
  PIXELFORMATDESCRIPTOR pfd2;
  int maxpixelformat=DescribePixelFormat(dc,1,pfd2.size,&pfd2);
  if (!maxpixelformat)
    Win32ErrorMsg("DescribePixelFormat failed, no valid formats:");
  int i;
  int pf=0;
  int bestscore=-0x7FFFFFFF;
  for (i=1; i<maxpixelformat; ++i)
  {
    if (!DescribePixelFormat(dc,i,pfd2.size,&pfd2))
      continue;
    int score=0;
    if (pfd2.cColorBits==pfd.cColorBits)
      score+=0x2000;
    else if (pfd2.cColorBits>pfd.cColorBits)
      score+=0x1800+pfd.cColorBits-pfd2.cColorBits;
    if (pfd2.cDepthBits==pfd.cDepthBits)
      score+=0x1000;
    else if (pfd2.cDepthBits>pfd.cDepthBits)
      score+=0x800+pfd.cDepthBits-pfd2.cDepthBits;
    else
      score-=0x8000;
    if (pfd2.cAccumBits==pfd.cAccumBits)
      score+=0x8000;
    else if (pfd2.cAccumBits>pfd.cAccumBits)
      score+=0x400+pfd.cAccumBits-pfd2.cAccumBits;
    else
      score-=0x400;
    if (pfd2.cStencilBits==pfd.cStencilBits)
      score+=0x800;
    else if (pfd2.cStencilBits>pfd.cStencilBits)
      score+=0x400+pfd.cStencilBits-pfd2.cStencilBits;
    else
      score-=0x400;
    if (pfd2.cAuxBuffers==pfd.cAuxBuffers)
      score+=0x800;
    else if (pfd2.cAuxBuffers>pfd.cAuxBuffers)
      score+=0x400+pfd.cAuxBuffers-pfd2.cAuxBuffers;
    else
      score-=0x400;
    if (!(pfd2.dwFlags&PFD_GENERIC_FORMAT))
      score+=0x1000;
    else if (pfd2.dwFlags&PFD_GENERIC_ACCELERATED)
      score+=0x800;
    if (pfd.dwFlags&PFD_DRAW_TO_BITMAP)
    {
      if (!(pfd2.dwFlags&PFD_DRAW_TO_BITMAP) ||
          (pfd2.dwFlags&PFD_DOUBLEBUFFER))
        score=0;
      if (pfd2.dwFlags&PFD_DRAW_TO_WINDOW)
        score-=0xC00;
    }
    if (pfd.dwFlags&PFD_DRAW_TO_WINDOW)
    {
      if (!(pfd2.dwFlags&PFD_DRAW_TO_WINDOW))
        score=0;
      if (pfd2.dwFlags&PFD_DRAW_TO_BITMAP)
        score-=0xC00;
    }
    if (pfd.dwFlags&PFD_SUPPORT_OPENGL)
      if (!(pfd2.dwFlags&PFD_SUPPORT_OPENGL))
        score=0;
    if (pfd.dwFlags&PFD_SUPPORT_GDI)
      if (!(pfd2.dwFlags&PFD_SUPPORT_GDI))
        score=0;
    if (pfd.dwFlags&PFD_DOUBLEBUFFER)
      if (!(pfd2.dwFlags&PFD_DOUBLEBUFFER_DONTCARE))
        if (!(pfd2.dwFlags&PFD_DOUBLEBUFFER))
          score=0;
        else
          score+=0x200;
    if (pfd2.iPixelType!=pfd.iPixelType)
      score=-0x1000;
    if (pfd2.iLayerType!=pfd.iLayerType)
      score=-0x800;

    // printf("Pixel format %d, score=0x%lX\n",i,score);
    if (score>bestscore)
    {
      pf=i;
      bestscore=score;
      // printf("  New best!\n");
    }
    // DumpPixelFormat(pfd2);
  }
  printf("Best pixel format is %d, score=0x%lX\n",pf,bestscore);
  if (pf)
    return pf;
  // printf("no matches, trying ChoosePixelFormat\n");
  pf = ChoosePixelFormat(dc, &pfd);
  if (pf == 0)
  {
    Win32ErrorMsg("ChoosePixelFormat failed.");
  }
  return pf;
}

bit SetupPixelFormat(HDC dc, int flags)
{
  PIXELFORMATDESCRIPTOR pfd;
  pfd.nSize=PIXELFORMATDESCRIPTOR.size;
  pfd.nVersion=1;
  pfd.dwFlags=flags | PFD_SUPPORT_OPENGL;
  pfd.iPixelType=PFD_TYPE_RGBA;       // color type
  pfd.cColorBits=bitspixel;           // prefered color depth
  pfd.cRedBits=pfd.cRedShift=pfd.cGreenBits=pfd.cGreenShift=
  pfd.cBlueBits=pfd.cBlueShift=pfd.cAlphaBits=pfd.cAlphaShift=0;
  pfd.cAccumBits=pfd.cAccumRedBits=pfd.cAccumGreenBits=pfd.cAccumBlueBits=pfd.cAccumAlphaBits=0;
  pfd.cDepthBits=16;                  // depth buffer
  pfd.cStencilBits=pfd.cAuxBuffers=0; // no stencil bits/auxiliary buffers
  pfd.iLayerType=PFD_MAIN_PLANE;      // main layer
  pfd.bReserved=0;
  pfd.dwLayerMask=pfd.dwVisibleMask=pfd.dwDamageMask=0; // no layer, visible, damage masks

  int pixelFormat;

  printf("ChoosePixelFormat\n");
  pixelFormat = ChoosePixelFormat2(dc, pfd);
  if (pixelFormat == 0)
  {
    Win32ErrorMsg("ChoosePixelFormat failed.");
  }

  // try to clean up after ChoosePixelFormat
  pfd.cDepthBits=16;                  // depth buffer
  pfd.cAccumBits=pfd.cAccumRedBits=pfd.cAccumGreenBits=pfd.cAccumBlueBits=pfd.cAccumAlphaBits=0;
  pfd.cStencilBits=pfd.cAuxBuffers=0; // no stencil bits/auxiliary buffers

  bit softwarerendering=false;
  if ((pfd.dwFlags & PFD_GENERIC_FORMAT) &&
      !(pfd.dwFlags & PFD_GENERIC_ACCELERATED))
  {
    softwarerendering=true;
    printf("Didn't get hardware acceleration.\n");
  }

  printf("SetPixelFormat\n");
  if (!SetPixelFormat(dc, pixelFormat, &pfd))
  {
    printf("SetPixelFormat failed.\n");
  }

  return !softwarerendering;
}

bit InitGL()
{
  printf("Initializing driver\n");
  HDC hdc=GetDC(hwnd);
  bitspixel = GetDeviceCaps(hdc,BITSPIXEL);
  dev = null;
  if      (bitspixel<  8) bitspixel= 8;
  else if (bitspixel==15) bitspixel=16;
  else if (bitspixel> 24) bitspixel=24;

  printf("Setting Pixel Format\n");
  bit gothardware=SetupPixelFormat(hdc, PFD_DOUBLEBUFFER | // support double-buffering
                        PFD_DRAW_TO_WINDOW);
  printf("Creating GL context\n");
  dev = wglCreateContext(hdc);
  if (!dev)
  {
    Win32ErrorMsg("Can't create context.");
  }
  printf("Created GL context\n");
  if (!wglMakeCurrent(hdc, dev))
  {
    Win32ErrorMsg("Can't make current.");
  }
  CheckError();

  printf("VENDOR: %s\n",glGetString(GL_VENDOR));
  printf("RENDERER: %s\n",glGetString(GL_RENDERER));
  printf("VERSION: %s\n",glGetString(GL_VERSION));

  SetupProjection();
  SetupCamera();
  // glViewport(0,0,w,h);

  printf("Initialized driver\n");
  return true;
}

void DoneGL()
{
  printf("Deinitializing driver...\n");
  CheckError();
  glFinish();
  CheckError();
  if (dev)
  {
    wglMakeCurrent(null, null);
    wglDeleteContext(dev);
    printf("GL context shut down\n");
    dev=null;
  }
  printf("Deinitialized driver\n");
}

bit ProcessEvents()
{
  MSG msg;
  while (PeekMessage(&msg, cast(HWND) null, 0, 0, PM_NOREMOVE))
  {
    if (!GetMessage(&msg, cast(HWND) null, 0, 0))
      return false;
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  return true;
}

extern(Windows)
int WindowProc(HWND mhwnd, uint msg, int wp, int lp)
{
  if (mhwnd==hwnd || msg==WM_CREATE)
  switch (msg)
  {
    case WM_CLOSE:
      PostQuitMessage(0);
      return 0; // don't allow DefWndProc to post WM_DESTROY
    case WM_KEYUP:
      if (VK_ESCAPE == wp)
      {
        PostMessage(mhwnd,WM_CLOSE,0,0);
        return 0;
      }
      break;
    case WM_SYSCOMMAND:
    {
      // disable screen saver
      if ((wp & 0xFFF0) == SC_SCREENSAVE ||    // disable screensaver
          (wp & 0xFFF0) == SC_MONITORPOWER ||  // disable monitor suspend
          (wp & 0xFFF0) == SC_KEYMENU)         // disable ALT key menu
        return 0;
      break;
    }
    case WM_SYSCHAR:
      if (wp==VK_F10) // disable app menu
        return 0;
      break;
    case WM_PAINT:
      ValidateRect(hwnd,null);
      return 0;  // we'll paint ourselves elsewhere
    case WM_SIZE:
      if (wp!=SIZE_MINIMIZED)
        Resized(LOWORD(lp),HIWORD(lp));
      break;
    default:
      break;
  }

  return DefWindowProc(mhwnd, msg, wp, lp);
}

bit InitWindow()
{
  printf("Creating window\n");
  const char* classname="TestGLClass";
  const char* windowtitle="TestGL";
  DWORD style= WS_VISIBLE |
      WS_CLIPCHILDREN | WS_CLIPSIBLINGS |
      WS_SYSMENU | WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
      WS_POPUPWINDOW | WS_THICKFRAME | WS_CAPTION;
  WNDCLASS wc;
  wc.style =
    CS_OWNDC |
    CS_DBLCLKS |
    CS_BYTEALIGNCLIENT |
    CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = (WNDPROC)&WindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hinst;
  wc.hIcon = LoadIconA(null, IDI_APPLICATION);
  wc.hCursor = LoadCursorA(null, IDC_CROSS);
  wc.hbrBackground = null;
  // Stock brushes cause some spurious GDI32 exceptions
  // wc.hbrBackground = GetStockObject(NULL_BRUSH);
  wc.lpszMenuName = null;
  wc.lpszClassName = classname;
  if (!RegisterClass(&wc))
  {
    assert(false);
    return false;
  }
  RECT r;
  r.left=r.top=0;
  r.right=w; r.bottom=h;
  AdjustWindowRect(&r,style,false); // assume no menu bar
  hwnd = CreateWindowEx(0,classname,windowtitle,style,
      0,0,r.right-r.left,r.bottom-r.top,
      null,null,hinst,null);
  if (!hwnd)
  {
    assert(false);
    return false;
  }
  return true;
}

void DoneWindow()
{
  if (hwnd)
  {
    printf("Destroying window\n");
    DestroyWindow(hwnd);
    hwnd=null;
  }
}

int main(char[][] args)
{
  hinst=GetModuleHandle(null);
  w=640; h=480;
  if (InitWindow())
  {
    if (InitGL())
    {
      if (InitDemo())
      {
        printf("running\n");
        int starttime,endtime;
        int nframes;
        time.time(&starttime);
        do
        {
          ++nframes;
          InitRender();
          Render();
          DoneRender();
        } while (ProcessEvents());
        time.time(&endtime);
        printf("did %d frames in %d seconds, that's %f FPS\n",nframes,endtime-starttime,cast(double)nframes/(endtime-starttime));
        printf("closing\n");
        DoneDemo();
      }
      DoneGL();
    }
    DoneWindow();
  }
  return 0;
}

